@@ -36,26 +36,25 @@
 #include <sys/ioctl.h>
 
 #include <bluetooth/bluetooth.h>
+#include <bluetooth/hci.h>
 #include <bluetooth/l2cap.h>
-#include <bluetooth/sdp.h>
-#include <bluetooth/sdp_lib.h>
 
 #include <glib.h>
 #include <dbus/dbus.h>
 #include <gdbus.h>
 
-#include "hcid.h"
+#include "logging.h"
+
 #include "manager.h"
 #include "adapter.h"
 #include "dbus-hci.h"
-#include "dbus-database.h"
 #include "dbus-common.h"
 
 #define BLUEZ_NAME "org.bluez"
 
 #define RECONNECT_RETRY_TIMEOUT	5000
 
-int l2raw_connect(const char *local, const bdaddr_t *remote)
+int l2raw_connect(const bdaddr_t *src, const bdaddr_t *dst)
 {
 	struct sockaddr_l2 addr;
 	long arg;
@@ -69,7 +68,7 @@ int l2raw_connect(const char *local, const bdaddr_t *remote)
 
 	memset(&addr, 0, sizeof(addr));
 	addr.l2_family = AF_BLUETOOTH;
-	str2ba(local, &addr.l2_bdaddr);
+	bacpy(&addr.l2_bdaddr, src);
 
 	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
 		error("Can't bind socket: %s (%d)", strerror(errno), errno);
@@ -90,7 +89,7 @@ int l2raw_connect(const char *local, const bdaddr_t *remote)
 
 	memset(&addr, 0, sizeof(addr));
 	addr.l2_family = AF_BLUETOOTH;
-	bacpy(&addr.l2_bdaddr, remote);
+	bacpy(&addr.l2_bdaddr, dst);
 
 	if (connect(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
 		if (errno == EAGAIN || errno == EINPROGRESS)
@@ -182,7 +181,7 @@ void hcid_dbus_unregister(void)
 
 	for (i = 0; children[i]; i++) {
 		char path[MAX_PATH_LENGTH];
-		struct adapter *adapter;
+		struct btd_adapter *adapter;
 
 		if (children[i][0] != 'h')
 			continue;
