@@ -60,6 +60,64 @@
 
 static DBusConnection *connection = NULL;
 
+const char *class_to_icon(uint32_t class)
+{
+	switch ((class & 0x1f00) >> 8) {
+	case 0x01:
+		return "computer";
+	case 0x02:
+		switch ((class & 0xfc) >> 2) {
+		case 0x01:
+		case 0x02:
+		case 0x03:
+		case 0x05:
+			return "phone";
+		case 0x04:
+			return "modem";
+		}
+		break;
+	case 0x03:
+		return "network-wireless";
+	case 0x04:
+		switch ((class & 0xfc) >> 2) {
+		case 0x01:
+		case 0x02:
+			return "audio-card";	/* Headset */
+		case 0x06:
+			return "audio-card";	/* Headphone */
+		}
+		break;
+	case 0x05:
+		switch ((class & 0xc0) >> 6) {
+		case 0x00:
+			switch ((class & 0x1e) >> 2) {
+			case 0x01:
+			case 0x02:
+				return "input-gaming";
+			}
+			break;
+		case 0x01:
+			return "input-keyboard";
+		case 0x02:
+			switch ((class & 0x1e) >> 2) {
+			case 0x05:
+				return "input-tablet";
+			default:
+				return "input-mouse";
+			}
+		}
+		break;
+	case 0x06:
+		if (class & 0x80)
+			return "printer";
+		if (class & 0x20)
+			return "camera-photo";
+		break;
+	}
+
+	return NULL;
+}
+
 DBusMessage *new_authentication_return(DBusMessage *msg, uint8_t status)
 {
 	switch (status) {
@@ -518,7 +576,7 @@ void hcid_dbus_bonding_process_complete(bdaddr_t *local, bdaddr_t *peer,
 
 	bonding = adapter_get_bonding_info(adapter);
 
-	if (bacmp(&bonding->bdaddr, peer))
+	if (bonding && bacmp(&bonding->bdaddr, peer))
 		bonding = NULL;
 
 	if (status == 0) {
@@ -649,7 +707,7 @@ static int found_device_req_name(struct btd_adapter *adapter)
 		}
 
 		error("Remote name request failed with status 0x%02x",
-			rp.status);
+								rp.status);
 
 		/* if failed, request the next element */
 		/* remove the element from the list */
@@ -825,13 +883,12 @@ void hcid_dbus_inquiry_result(bdaddr_t *local, bdaddr_t *peer, uint32_t class,
 {
 	char filename[PATH_MAX + 1];
 	struct btd_adapter *adapter;
-	char local_addr[18], peer_addr[18], *name, *tmp_name;
-	const char *paddr = peer_addr;
+	char local_addr[18], peer_addr[18], *alias, *name, *tmp_name;
+	const char *path, *icon, *paddr = peer_addr;
 	struct remote_dev_info *dev, match;
 	dbus_int16_t tmp_rssi = rssi;
 	uint8_t name_type = 0x00;
 	name_status_t name_status;
-	const gchar *path;
 	int state;
 
 	ba2str(local, local_addr);
@@ -876,9 +933,20 @@ void hcid_dbus_inquiry_result(bdaddr_t *local, bdaddr_t *peer, uint32_t class,
 	else
 		name_status = NAME_NOT_REQUIRED;
 
+	create_name(filename, PATH_MAX, STORAGEDIR, local_addr, "aliases");
+	alias = textfile_get(filename, peer_addr);
+
 	create_name(filename, PATH_MAX, STORAGEDIR, local_addr, "names");
 	name = textfile_get(filename, peer_addr);
 
+	if (!alias) {
+		if (!name) {
+			alias = g_strdup(peer_addr);
+			g_strdelimit(alias, ":", '-');
+		} else
+			alias = g_strdup(name);
+	}
+
 	tmp_name = extract_eir_name(data, &name_type);
 	if (tmp_name) {
 		if (name_type == 0x09) {
@@ -898,6 +966,7 @@ void hcid_dbus_inquiry_result(bdaddr_t *local, bdaddr_t *peer, uint32_t class,
 	}
 
 	path = adapter_get_path(adapter);
+	icon = class_to_icon(class);
 
 	if (name) {
 		if (name_type != 0x08)
@@ -906,19 +975,23 @@ void hcid_dbus_inquiry_result(bdaddr_t *local, bdaddr_t *peer, uint32_t class,
 		emit_device_found(path, paddr,
 				"Address", DBUS_TYPE_STRING, &paddr,
 				"Class", DBUS_TYPE_UINT32, &class,
+				"Icon", DBUS_TYPE_STRING, &icon,
 				"RSSI", DBUS_TYPE_INT16, &tmp_rssi,
 				"Name", DBUS_TYPE_STRING, &name,
-				NULL);
+				"Alias", DBUS_TYPE_STRING, &alias, NULL);
 
 		g_free(name);
 	} else {
 		emit_device_found(path, paddr,
 				"Address", DBUS_TYPE_STRING, &paddr,
 				"Class", DBUS_TYPE_UINT32, &class,
+				"Icon", DBUS_TYPE_STRING, &icon,
 				"RSSI", DBUS_TYPE_INT16, &tmp_rssi,
-				NULL);
+				"Alias", DBUS_TYPE_STRING, &alias, NULL);
 	}
 
+	g_free(alias);
+
 	/* add in the list to track name sent/pending */
 	adapter_add_found_device(adapter, peer, rssi, name_status);
 }
@@ -960,8 +1033,7 @@ void hcid_dbus_remote_name(bdaddr_t *local, bdaddr_t *peer, uint8_t status,
 				char *name)
 {
 	struct btd_adapter *adapter;
-	char peer_addr[18];
-	const char *paddr = peer_addr;
+	char srcaddr[18], dstaddr[18];
 	const gchar *dev_path;
 	int state;
 
@@ -971,19 +1043,29 @@ void hcid_dbus_remote_name(bdaddr_t *local, bdaddr_t *peer, uint8_t status,
 		return;
 	}
 
-	ba2str(peer, peer_addr);
+	ba2str(local, srcaddr);
+	ba2str(peer, dstaddr);
 
 	if (!status) {
 		struct btd_device *device;
 
-		device = adapter_find_device(adapter, paddr);
+		device = adapter_find_device(adapter, dstaddr);
 		if (device) {
+			char alias[248];
 
 			dev_path = device_get_path(device);
 
 			dbus_connection_emit_property_changed(connection,
 						dev_path, DEVICE_INTERFACE,
 						"Name", DBUS_TYPE_STRING, &name);
+
+			if (read_device_alias(srcaddr, dstaddr,
+						alias, sizeof(alias)) < 1) {
+
+				dbus_connection_emit_property_changed(connection,
+						dev_path, DEVICE_INTERFACE,
+						"Alias", DBUS_TYPE_STRING, &name);
+			}
 		}
 	}
 
@@ -1143,7 +1225,6 @@ void hcid_dbus_disconn_complete(bdaddr_t *local, uint8_t status,
 int set_limited_discoverable(int dd, const uint8_t *cls, gboolean limited)
 {
 	uint32_t dev_class;
-	int err;
 	int num = (limited ? 2 : 1);
 	uint8_t lap[] = { 0x33, 0x8b, 0x9e, 0x00, 0x8b, 0x9e };
 	/*
@@ -1151,7 +1232,7 @@ int set_limited_discoverable(int dd, const uint8_t *cls, gboolean limited)
 	 * 2: giac + liac
 	 */
 	if (hci_write_current_iac_lap(dd, num, lap, HCI_REQ_TIMEOUT) < 0) {
-		err = errno;
+		int err = errno;
 		error("Can't write current IAC LAP: %s(%d)",
 				strerror(err), err);
 		return -err;
@@ -1170,7 +1251,7 @@ int set_limited_discoverable(int dd, const uint8_t *cls, gboolean limited)
 	}
 
 	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
-		err = errno;
+		int err = errno;
 		error("Can't write class of device: %s (%d)",
 							strerror(err), err);
 		return -err;
@@ -1182,7 +1263,6 @@ int set_limited_discoverable(int dd, const uint8_t *cls, gboolean limited)
 int set_service_classes(int dd, const uint8_t *cls, uint8_t value)
 {
 	uint32_t dev_class;
-	int err;
 
 	if (cls[2] == value)
 		return 0; /* Already set */
@@ -1190,7 +1270,58 @@ int set_service_classes(int dd, const uint8_t *cls, uint8_t value)
 	dev_class = (value << 16) | (cls[1] << 8) | cls[0];
 
 	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
-		err = errno;
+		int err = errno;
+		error("Can't write class of device: %s (%d)",
+							strerror(err), err);
+		return -err;
+	}
+
+	return 0;
+}
+
+int set_major_class(int dd, const uint8_t *cls, uint8_t major)
+{
+	uint32_t dev_class;
+
+	dev_class = (cls[2] << 16) | ((cls[1] & 0x20) << 8) |
+						((major & 0xdf) << 8) | 0x00;
+
+	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
+		int err = errno;
+		error("Can't write class of device: %s (%d)",
+							strerror(err), err);
+		return -err;
+	}
+
+	return 0;
+}
+
+int set_minor_class(int dd, const uint8_t *cls, uint8_t minor)
+{
+	uint32_t dev_class;
+
+	dev_class = (cls[2] << 16) | (cls[1] << 8) | minor;
+
+	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
+		int err = errno;
+		error("Can't write class of device: %s (%d)",
+							strerror(err), err);
+		return -err;
+	}
+
+	return 0;
+}
+
+int set_major_and_minor_class(int dd, const uint8_t *cls,
+						uint8_t major, uint8_t minor)
+{
+	uint32_t dev_class;
+
+	dev_class = (cls[2] << 16) | ((cls[1] & 0x20) << 8) |
+						((major & 0xdf) << 8) | minor;
+
+	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
+		int err = errno;
 		error("Can't write class of device: %s (%d)",
 							strerror(err), err);
 		return -err;
@@ -1325,8 +1456,8 @@ void hcid_dbus_write_class_complete(bdaddr_t *local)
 		return;
 	}
 
-	write_local_class(local, cls);
 	adapter_set_class(adapter, cls);
+	write_local_class(local, cls);
 
 	hci_close_dev(dd);
 }
@@ -1428,7 +1559,7 @@ int hcid_dbus_get_io_cap(bdaddr_t *local, bdaddr_t *remote,
 }
 
 int hcid_dbus_set_io_cap(bdaddr_t *local, bdaddr_t *remote,
-                                                uint8_t cap, uint8_t auth)
+						uint8_t cap, uint8_t auth)
 {
 	struct btd_adapter *adapter;
 	struct btd_device *device;
