@@ -30,6 +30,7 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <sys/stat.h>
+#include <errno.h>
 
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/hci.h>
@@ -57,6 +58,7 @@
 
 #define DEFAULT_XML_BUF_SIZE	1024
 #define DISCONNECT_TIMER	2
+#define DISCOVERY_TIMER		2000
 
 struct btd_driver_data {
 	struct btd_device_driver *driver;
@@ -64,7 +66,7 @@ struct btd_driver_data {
 };
 
 struct btd_device {
-	gchar		*address;
+	bdaddr_t	bdaddr;
 	gchar		*path;
 	struct btd_adapter	*adapter;
 	GSList		*uuids;
@@ -75,10 +77,16 @@ struct btd_device {
 	int		discov_active;		/* Service discovery active */
 	char		*discov_requestor;	/* discovery requestor unique name */
 	guint		discov_listener;
+	guint		discov_timer;
 
 	/* For Secure Simple Pairing */
 	uint8_t		cap;
 	uint8_t		auth;
+
+	gboolean	connected;
+
+	/* Whether were creating a security mode 3 connection */
+	gboolean	secmode3;
 };
 
 struct browse_req {
@@ -126,7 +134,6 @@ static void device_free(gpointer user_data)
 	if (device->disconn_timer)
 		g_source_remove(device->disconn_timer);
 
-	g_free(device->address);
 	g_free(device->path);
 	g_free(device);
 }
@@ -134,16 +141,18 @@ static void device_free(gpointer user_data)
 static gboolean device_is_paired(struct btd_device *device)
 {
 	struct btd_adapter *adapter = device->adapter;
-	char filename[PATH_MAX + 1], *str, addr[18];
+	char filename[PATH_MAX + 1], *str;
+	char srcaddr[18], dstaddr[18];
 	gboolean ret;
-	bdaddr_t bdaddr;
+	bdaddr_t src;
 
-	adapter_get_address(adapter, &bdaddr);
-	ba2str(&bdaddr, addr);
+	adapter_get_address(adapter, &src);
+	ba2str(&src, srcaddr);
+	ba2str(&device->bdaddr, dstaddr);
 
 	create_name(filename, PATH_MAX, STORAGEDIR,
-			addr, "linkkeys");
-	str = textfile_caseget(filename, device->address);
+			srcaddr, "linkkeys");
+	str = textfile_caseget(filename, dstaddr);
 	ret = str ? TRUE : FALSE;
 	g_free(str);
 
@@ -158,8 +167,8 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	DBusMessage *reply;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
-	bdaddr_t src, dst;
-	char name[248], src_addr[18];
+	bdaddr_t src;
+	char name[248], srcaddr[18], dstaddr[18];
 	char **uuids;
 	const char *ptr;
 	dbus_bool_t boolean;
@@ -168,6 +177,8 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	GSList *l;
 	struct active_conn_info *dev;
 
+	ba2str(&device->bdaddr, dstaddr);
+
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
 		return NULL;
@@ -180,69 +191,79 @@ static DBusMessage *get_properties(DBusConnection *conn,
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
 	/* Address */
-	dbus_message_iter_append_dict_entry(&dict, "Address", DBUS_TYPE_STRING,
-			&device->address);
+	ptr = dstaddr;
+	dbus_message_iter_append_dict_entry(&dict, "Address",
+						DBUS_TYPE_STRING, &ptr);
 
 	/* Name */
 	ptr = NULL;
 	memset(name, 0, sizeof(name));
 	adapter_get_address(adapter, &src);
-	ba2str(&src, src_addr);
+	ba2str(&src, srcaddr);
 
-	if (read_device_name(src_addr, device->address, name) == 0) {
+	if (read_device_name(srcaddr, dstaddr, name) == 0) {
 		ptr = name;
 		dbus_message_iter_append_dict_entry(&dict, "Name",
-				DBUS_TYPE_STRING, &ptr);
+						DBUS_TYPE_STRING, &ptr);
 	}
 
-	if (read_device_alias(src_addr, device->address, name, sizeof(name)) > 0)
+	/* Alias (fallback to name or address) */
+	if (read_device_alias(srcaddr, dstaddr, name, sizeof(name)) < 1) {
+		if (!ptr) {
+			g_strdelimit(dstaddr, ":", '-');
+			ptr = dstaddr;
+		}
+	} else
 		ptr = name;
 
-	/* Alias: use Name if Alias doesn't exist */
 	if (ptr)
 		dbus_message_iter_append_dict_entry(&dict, "Alias",
-				DBUS_TYPE_STRING, &ptr);
-
-	str2ba(device->address, &dst);
+						DBUS_TYPE_STRING, &ptr);
 
 	/* Class */
-	if (read_remote_class(&src, &dst, &class) == 0) {
+	if (read_remote_class(&src, &device->bdaddr, &class) == 0) {
+		const char *icon = class_to_icon(class);
+
 		dbus_message_iter_append_dict_entry(&dict, "Class",
-				DBUS_TYPE_UINT32, &class);
+						DBUS_TYPE_UINT32, &class);
+
+		if (icon)
+			dbus_message_iter_append_dict_entry(&dict, "Icon",
+						DBUS_TYPE_STRING, &icon);
 	}
 
 	/* Paired */
 	boolean = device_is_paired(device);
 	dbus_message_iter_append_dict_entry(&dict, "Paired",
-			DBUS_TYPE_BOOLEAN, &boolean);
+						DBUS_TYPE_BOOLEAN, &boolean);
 
 	/* Trusted */
-	boolean = read_trust(&src, device->address, GLOBAL_TRUST);
+	boolean = read_trust(&src, dstaddr, GLOBAL_TRUST);
 	dbus_message_iter_append_dict_entry(&dict, "Trusted",
-			DBUS_TYPE_BOOLEAN, &boolean);
+						DBUS_TYPE_BOOLEAN, &boolean);
 
 	/* Connected */
-	dev = adapter_search_active_conn_by_bdaddr(adapter, &dst);
+	dev = adapter_search_active_conn_by_bdaddr(adapter, &device->bdaddr);
 	if (dev)
 		boolean = TRUE;
 	else
 		boolean = FALSE;
 
 	dbus_message_iter_append_dict_entry(&dict, "Connected",
-			DBUS_TYPE_BOOLEAN, &boolean);
+						DBUS_TYPE_BOOLEAN, &boolean);
 
 	/* UUIDs */
 	uuids = g_new0(char *, g_slist_length(device->uuids) + 1);
 	for (i = 0, l = device->uuids; l; l = l->next, i++)
 		uuids[i] = l->data;
 	dbus_message_iter_append_dict_entry(&dict, "UUIDs",
-			DBUS_TYPE_ARRAY, &uuids);
+						DBUS_TYPE_ARRAY, &uuids);
 	g_free(uuids);
 
 	/* Adapter */
 	ptr = adapter_get_path(adapter);
 	dbus_message_iter_append_dict_entry(&dict, "Adapter",
-			DBUS_TYPE_OBJECT_PATH, &ptr);
+						DBUS_TYPE_OBJECT_PATH, &ptr);
 
 	dbus_message_iter_close_container(&iter, &dict);
 
@@ -254,15 +275,16 @@ static DBusMessage *set_alias(DBusConnection *conn, DBusMessage *msg,
 {
 	struct btd_device *device = data;
 	struct btd_adapter *adapter = device->adapter;
-	char addr[18];
-	bdaddr_t bdaddr;
+	char srcaddr[18], dstaddr[18];
+	bdaddr_t src;
 	int err;
 
-	adapter_get_address(adapter, &bdaddr);
-	ba2str(&bdaddr, addr);
+	adapter_get_address(adapter, &src);
+	ba2str(&src, srcaddr);
+	ba2str(&device->bdaddr, dstaddr);
 
-	 /* Remove alias if empty string */
-	err = write_device_alias(addr, device->address,
+	/* Remove alias if empty string */
+	err = write_device_alias(srcaddr, dstaddr,
 			g_str_equal(alias, "") ? NULL : alias);
 	if (err < 0)
 		return g_dbus_create_error(msg,
@@ -281,14 +303,15 @@ static DBusMessage *set_trust(DBusConnection *conn, DBusMessage *msg,
 {
 	struct btd_device *device = data;
 	struct btd_adapter *adapter = device->adapter;
-	char addr[18];
-	bdaddr_t bdaddr;
+	char srcaddr[18], dstaddr[18];
+	bdaddr_t src;
 
 
-	adapter_get_address(adapter, &bdaddr);
-	ba2str(&bdaddr, addr);
+	adapter_get_address(adapter, &src);
+	ba2str(&src, srcaddr);
+	ba2str(&device->bdaddr, dstaddr);
 
-	write_trust(addr, device->address, GLOBAL_TRUST, value);
+	write_trust(srcaddr, dstaddr, GLOBAL_TRUST, value);
 
 	dbus_connection_emit_property_changed(conn, dbus_message_get_path(msg),
 					DEVICE_INTERFACE, "Trusted",
@@ -349,15 +372,13 @@ static void discover_services_req_exit(void *user_data)
 {
 	struct btd_device *device = user_data;
 	struct btd_adapter *adapter = device->adapter;
-	bdaddr_t src, dst;
+	bdaddr_t src;
 
 	adapter_get_address(adapter, &src);
 
 	debug("DiscoverDevices requestor exited");
 
-	str2ba(device->address, &dst);
-
-	bt_cancel_discovery(&src, &dst);
+	bt_cancel_discovery(&src, &device->bdaddr);
 }
 
 static DBusMessage *discover_services(DBusConnection *conn,
@@ -402,7 +423,7 @@ static DBusMessage *cancel_discover(DBusConnection *conn,
 {
 	struct btd_device *device = user_data;
 	struct btd_adapter *adapter = device->adapter;
-	bdaddr_t src, dst;
+	bdaddr_t src;
 
 	adapter_get_address(adapter, &src);
 
@@ -418,9 +439,7 @@ static DBusMessage *cancel_discover(DBusConnection *conn,
 				ERROR_INTERFACE ".NotAuthorized",
 				"Not Authorized");
 
-	str2ba(device->address, &dst);
-
-	if (bt_cancel_discovery(&src, &dst) < 0)
+	if (bt_cancel_discovery(&src, &device->bdaddr) < 0)
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".Failed",
 				"No pending discover");
@@ -433,15 +452,13 @@ static gboolean disconnect_timeout(gpointer user_data)
 	struct btd_device *device = user_data;
 	struct active_conn_info *ci;
 	disconnect_cp cp;
-	bdaddr_t bda;
 	int dd;
 	uint16_t dev_id = adapter_get_dev_id(device->adapter);
 
 	device->disconn_timer = 0;
 
-	str2ba(device->address, &bda);
-
-	ci = adapter_search_active_conn_by_bdaddr(device->adapter, &bda);
+	ci = adapter_search_active_conn_by_bdaddr(device->adapter,
+							&device->bdaddr);
 
 	if (!ci)
 		return FALSE;
@@ -467,12 +484,10 @@ static DBusMessage *disconnect(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
 	struct btd_device *device = user_data;
-	bdaddr_t bda;
 	struct active_conn_info *dev;
 
-	str2ba(device->address, &bda);
-
-	dev = adapter_search_active_conn_by_bdaddr(device->adapter, &bda);
+	dev = adapter_search_active_conn_by_bdaddr(device->adapter,
+							&device->bdaddr);
 
 	if (!dev)
 		return g_dbus_create_error(msg,
@@ -505,6 +520,43 @@ static GDBusSignalTable device_signals[] = {
 	{ }
 };
 
+gboolean device_get_connected(struct btd_device *device)
+{
+	return device->connected;
+}
+
+void device_set_connected(DBusConnection *conn, struct btd_device *device,
+			gboolean connected)
+{
+	device->connected = connected;
+
+	if (!connected && device->discov_timer) {
+		g_source_remove(device->discov_timer);
+		device->discov_timer = 0;
+	}
+
+	dbus_connection_emit_property_changed(conn, device->path,
+						DEVICE_INTERFACE,
+						"Connected", DBUS_TYPE_BOOLEAN,
+						&connected);
+
+	if (connected && device->secmode3) {
+		struct btd_adapter *adapter = device_get_adapter(device);
+		bdaddr_t sba;
+
+		adapter_get_address(adapter, &sba);
+
+		device->secmode3 = FALSE;
+
+		hcid_dbus_bonding_process_complete(&sba, &device->bdaddr, 0);
+	}
+}
+
+void device_set_secmode3_conn(struct btd_device *device, gboolean enable)
+{
+	device->secmode3 = enable;
+}
+
 struct btd_device *device_create(DBusConnection *conn, struct btd_adapter *adapter,
 					const gchar *address)
 {
@@ -531,7 +583,7 @@ struct btd_device *device_create(DBusConnection *conn, struct btd_adapter *adapt
 		return NULL;
 	}
 
-	device->address = g_strdup(address);
+	str2ba(address, &device->bdaddr);
 	device->adapter = adapter;
 
 	return device;
@@ -560,35 +612,10 @@ void device_remove(DBusConnection *conn, struct btd_device *device)
 
 gint device_address_cmp(struct btd_device *device, const gchar *address)
 {
-	return strcasecmp(device->address, address);
-}
+	char addr[18];
 
-sdp_record_t *get_record(sdp_list_t *recs, const char *uuid)
-{
-	sdp_list_t *seq;
-
-	for (seq = recs; seq; seq = seq->next) {
-		sdp_record_t *rec = (sdp_record_t *) seq->data;
-		sdp_list_t *svcclass = NULL;
-		char *uuid_str;
-
-		if (sdp_get_service_classes(rec, &svcclass) < 0)
-			continue;
-
-		/* Extract the uuid */
-		uuid_str = bt_uuid2string(svcclass->data);
-		if (!uuid_str)
-			continue;
-
-		if (!strcasecmp(uuid_str, uuid)) {
-			sdp_list_free(svcclass, free);
-			free(uuid_str);
-			return rec;
-		}
-		sdp_list_free(svcclass, free);
-		free(uuid_str);
-	}
-	return NULL;
+	ba2str(&device->bdaddr, addr);
+	return strcasecmp(addr, address);
 }
 
 void device_probe_drivers(struct btd_device *device, GSList *uuids, sdp_list_t *recs)
@@ -610,7 +637,7 @@ void device_probe_drivers(struct btd_device *device, GSList *uuids, sdp_list_t *
 					(GCompareFunc) strcasecmp))
 				continue;
 
-			rec = get_record(recs, *uuid);
+			rec = find_record_in_list(recs, *uuid);
 			if (!rec)
 				continue;
 
@@ -635,20 +662,27 @@ void device_probe_drivers(struct btd_device *device, GSList *uuids, sdp_list_t *
 		}
 	}
 
-	for (list = uuids; list; list = list->next)
+	for (list = uuids; list; list = list->next) {
+		GSList *l = g_slist_find_custom(device->uuids, list->data,
+							(GCompareFunc) strcmp);
+		if (l)
+			continue;
+
 		device->uuids = g_slist_insert_sorted(device->uuids,
-				list->data, (GCompareFunc) strcmp);
+					list->data, (GCompareFunc) strcmp);
+	}
 }
 
 void device_remove_drivers(struct btd_device *device, GSList *uuids, sdp_list_t *recs)
 {
 	struct btd_adapter *adapter = device_get_adapter(device);
 	GSList *list;
-	char src_addr[18];
+	char srcaddr[18], dstaddr[18];
 	bdaddr_t src;
 
 	adapter_get_address(adapter, &src);
-	ba2str(&src, src_addr);
+	ba2str(&src, srcaddr);
+	ba2str(&device->bdaddr, dstaddr);
 
 	debug("Remove drivers for %s", device->path);
 
@@ -670,11 +704,11 @@ void device_remove_drivers(struct btd_device *device, GSList *uuids, sdp_list_t
 
 			g_free(driver_data);
 
-			rec = get_record(recs, *uuid);
+			rec = find_record_in_list(recs, *uuid);
 			if (!rec)
 				continue;
 
-			delete_record(src_addr, device->address, rec->handle);
+			delete_record(srcaddr, dstaddr, rec->handle);
 		}
 	}
 
@@ -776,6 +810,7 @@ static void discover_device_reply(struct browse_req *req, sdp_list_t *recs)
 
 static void services_changed(struct browse_req *req)
 {
+	DBusConnection *conn = get_dbus_connection();
 	struct btd_device *device = req->device;
 	char **uuids;
 	GSList *l;
@@ -785,7 +820,7 @@ static void services_changed(struct browse_req *req)
 	for (i = 0, l = device->uuids; l; l = l->next, i++)
 		uuids[i] = l->data;
 
-	dbus_connection_emit_property_changed(req->conn, device->path,
+	dbus_connection_emit_property_changed(conn, device->path,
 					DEVICE_INTERFACE, "UUIDs",
 					DBUS_TYPE_ARRAY, &uuids);
 
@@ -805,11 +840,12 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 	struct btd_device *device = req->device;
 	struct btd_adapter *adapter = device_get_adapter(device);
 	sdp_list_t *seq;
-	char src_addr[18];
+	char srcaddr[18], dstaddr[18];
 	bdaddr_t src;
 
 	adapter_get_address(adapter, &src);
-	ba2str(&src, src_addr);
+	ba2str(&src, srcaddr);
+	ba2str(&device->bdaddr, dstaddr);
 
 	for (seq = recs; seq; seq = seq->next) {
 		sdp_record_t *rec = (sdp_record_t *) seq->data;
@@ -826,10 +862,33 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 
 		/* Extract the first element and skip the remainning */
 		uuid_str = bt_uuid2string(svcclass->data);
-		if (!uuid_str)
+		if (!uuid_str) {
+			sdp_list_free(svcclass, free);
 			continue;
+		}
 
-		/* Driver uuid found */
+		if (!strcasecmp(uuid_str, PNP_UUID)) {
+			uint16_t source, vendor, product, version;
+			sdp_data_t *pdlist;
+
+			pdlist = sdp_data_get(rec, SDP_ATTR_VENDOR_ID_SOURCE);
+			source = pdlist ? pdlist->val.uint16 : 0x0000;
+
+			pdlist = sdp_data_get(rec, SDP_ATTR_VENDOR_ID);
+			vendor = pdlist ? pdlist->val.uint16 : 0x0000;
+
+			pdlist = sdp_data_get(rec, SDP_ATTR_PRODUCT_ID);
+			product = pdlist ? pdlist->val.uint16 : 0x0000;
+
+			pdlist = sdp_data_get(rec, SDP_ATTR_VERSION);
+			version = pdlist ? pdlist->val.uint16 : 0x0000;
+
+			if (source || vendor || product || version)
+				store_device_id(srcaddr, dstaddr, source,
+						vendor, product, version);
+		}
+
+		/* Driver UUID found */
 		l = g_slist_find_custom(req->uuids, uuid_str,
 				(GCompareFunc) strcasecmp);
 		if (l) {
@@ -840,12 +899,13 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 		}
 
 		/* Check for duplicates */
-		if (sdp_list_find(req->records, rec, rec_cmp))
+		if (sdp_list_find(req->records, rec, rec_cmp)) {
+			g_free(uuid_str);
+			sdp_list_free(svcclass, free);
 			continue;
+		}
 
-		debug("New record found: %s", uuid_str);
-
-		store_record(src_addr, device->address, rec);
+		store_record(srcaddr, dstaddr, rec);
 
 		/* Copy record */
 		if (sdp_gen_record_pdu(rec, &pdu) == 0) {
@@ -859,13 +919,13 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 		}
 
 		l = g_slist_find_custom(device->uuids, uuid_str,
-				(GCompareFunc) strcmp);
+							(GCompareFunc) strcmp);
 		if (!l)
 			req->uuids_added = g_slist_append(req->uuids_added,
-					uuid_str);
+								uuid_str);
 		else {
 			req->uuids_removed = g_slist_remove(req->uuids_removed,
-					l->data);
+								l->data);
 			g_free(uuid_str);
 		}
 
@@ -876,19 +936,18 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 static void store(struct btd_device *device)
 {
 	struct btd_adapter *adapter = device->adapter;
-	bdaddr_t src, dst;
+	bdaddr_t src;
 	char *str;
 
 	adapter_get_address(adapter, &src);
-	str2ba(device->address, &dst);
 
 	if (!device->uuids) {
-		write_device_profiles(&src, &dst, "");
+		write_device_profiles(&src, &device->bdaddr, "");
 		return;
 	}
 
 	str = bt_list2string(device->uuids);
-	write_device_profiles(&src, &dst, str);
+	write_device_profiles(&src, &device->bdaddr, str);
 	g_free(str);
 }
 
@@ -923,20 +982,27 @@ static void search_cb(sdp_list_t *recs, int err, gpointer user_data)
 	services_changed(req);
 
 proceed:
-
 	/* Store the device's profiles in the filesystem */
 	store(device);
 
-	if (dbus_message_is_method_call(req->msg, DEVICE_INTERFACE,
-			"DiscoverServices")) {
-		discover_device_reply(req, req->records);
-		goto cleanup;
-	}
+	if (req->msg) {
+		if (dbus_message_is_method_call(req->msg, DEVICE_INTERFACE,
+						"DiscoverServices")) {
+			discover_device_reply(req, req->records);
+			goto cleanup;
+		}
 
-	g_dbus_emit_signal(req->conn, dbus_message_get_path(req->msg),
+		g_dbus_emit_signal(req->conn, dbus_message_get_path(req->msg),
 				ADAPTER_INTERFACE, "DeviceCreated",
 				DBUS_TYPE_OBJECT_PATH, &device->path,
 				DBUS_TYPE_INVALID);
+	}
+
+	/* Update device list */
+	adapter_update_devices(device->adapter);
+
+	if (!req->msg)
+		goto cleanup;
 
 	/* Reply create device request */
 	reply = dbus_message_new_method_return(req->msg);
@@ -959,8 +1025,10 @@ cleanup:
 		device->discov_requestor = NULL;
 	}
 
-	dbus_message_unref(req->msg);
-	dbus_connection_unref(req->conn);
+	if (req->msg)
+		dbus_message_unref(req->msg);
+	if (req->conn)
+		dbus_connection_unref(req->conn);
 	g_slist_free(req->uuids_added);
 	g_slist_free(req->uuids_removed);
 	if (req->records)
@@ -974,22 +1042,23 @@ static void browse_cb(sdp_list_t *recs, int err, gpointer user_data)
 	struct browse_req *req = user_data;
 	struct btd_device *device = req->device;
 	struct btd_adapter *adapter = device->adapter;
-	bdaddr_t src, dst;
+	bdaddr_t src;
 	uuid_t uuid;
 
-	/* Public browsing successful or Single record requested */
-	if (err < 0 || (!req->search_uuid && recs))
+	/* If we have a valid response and req->search_uuid == 1, then
+	   public browsing was successful -- we don't need any more */
+	if (err < 0 || (req->search_uuid == 1 && recs))
 		goto done;
 
 	update_services(req, recs);
 
 	adapter_get_address(adapter, &src);
-	str2ba(device->address, &dst);
 
 	/* Search for mandatory uuids */
-	if (uuid_list[++req->search_uuid]) {
-		sdp_uuid16_create(&uuid, uuid_list[req->search_uuid]);
-		bt_search_service(&src, &dst, &uuid, browse_cb, user_data, NULL);
+	if (uuid_list[req->search_uuid]) {
+		sdp_uuid16_create(&uuid, uuid_list[req->search_uuid++]);
+		bt_search_service(&src, &device->bdaddr, &uuid,
+						browse_cb, user_data, NULL);
 		return;
 	}
 
@@ -1000,14 +1069,16 @@ static void browse_cb(sdp_list_t *recs, int err, gpointer user_data)
 		bt_string2uuid(&uuid, uuid_str);
 		req->uuids = g_slist_remove(req->uuids, uuid_str);
 		g_free(uuid_str);
-		bt_search_service(&src, &dst, &uuid, browse_cb, user_data, NULL);
+		bt_search_service(&src, &device->bdaddr, &uuid,
+						browse_cb, user_data, NULL);
 		return;
 	}
 
 	/* Search for l2cap uuid */
 	if (!req->records) {
 		sdp_uuid16_create(&uuid, L2CAP_UUID);
-		bt_search_service(&src, &dst, &uuid, search_cb, user_data, NULL);
+		bt_search_service(&src, &device->bdaddr, &uuid,
+						search_cb, user_data, NULL);
 		return;
 	}
 
@@ -1025,10 +1096,25 @@ static void init_browse(struct browse_req *req)
 
 		for (i = 0; driver->uuids[i]; i++) {
 			char *uuid;
+			int j;
 
+			/* Eliminate duplicates of UUIDs in uuid_list[]... */
+			if (strlen(driver->uuids[i]) == 36 && 
+			    !strncmp(driver->uuids[i], "0000", 4) &&
+			    !strcasecmp(driver->uuids[i] + 8, 
+					"-0000-1000-8000-00805F9B34FB")) {
+				uint16_t uuid16 = strtol(driver->uuids[i],
+							 NULL, 16);
+				for (j = 0; uuid_list[j]; j++) {
+					if (uuid16 == uuid_list[j])
+						continue;
+				}
+
+			}
+			/* ... and of UUIDs another driver already asked for */
 			if (g_slist_find_custom(req->uuids, driver->uuids[i],
 					(GCompareFunc) strcasecmp))
-				return;
+				continue;
 
 			uuid = g_strdup(driver->uuids[i]);
 			req->uuids = g_slist_append(req->uuids, uuid);
@@ -1045,38 +1131,45 @@ int device_browse(struct btd_device *device, DBusConnection *conn,
 {
 	struct btd_adapter *adapter = device->adapter;
 	struct browse_req *req;
-	bdaddr_t src, dst;
+	bdaddr_t src;
 	uuid_t uuid;
 	bt_callback_t cb;
 
+	if (device->discov_active)
+		return -EBUSY;
+
 	adapter_get_address(adapter, &src);
 
 	req = g_new0(struct browse_req, 1);
-	req->conn = dbus_connection_ref(conn);
-	req->msg = dbus_message_ref(msg);
+	if (conn && msg) {
+		req->conn = dbus_connection_ref(conn);
+		req->msg = dbus_message_ref(msg);
+	}
 	req->device = device;
 
-	str2ba(device->address, &dst);
-
 	if (search) {
 		memcpy(&uuid, search, sizeof(uuid_t));
 		cb = search_cb;
 	} else {
-		sdp_uuid16_create(&uuid, uuid_list[req->search_uuid]);
+		sdp_uuid16_create(&uuid, uuid_list[req->search_uuid++]);
 		init_browse(req);
 		cb = browse_cb;
 	}
 
 	device->discov_active = 1;
-	device->discov_requestor = g_strdup(dbus_message_get_sender(msg));
-	/* Track the request owner to cancel it
-	 * automatically if the owner exits */
-	device->discov_listener = g_dbus_add_disconnect_watch(conn,
+
+	if (msg) {
+		device->discov_requestor = g_strdup(dbus_message_get_sender(msg));
+		/* Track the request owner to cancel it
+		 * automatically if the owner exits */
+		device->discov_listener = g_dbus_add_disconnect_watch(conn,
 						dbus_message_get_sender(msg),
 						discover_services_req_exit,
 						device, NULL);
+	}
 
-	return bt_search_service(&src, &dst, &uuid, browse_cb, req, NULL);
+	return bt_search_service(&src, &device->bdaddr,
+					&uuid, cb, req, NULL);
 }
 
 struct btd_adapter *device_get_adapter(struct btd_device *device)
@@ -1089,7 +1182,7 @@ struct btd_adapter *device_get_adapter(struct btd_device *device)
 
 void device_get_address(struct btd_device *device, bdaddr_t *bdaddr)
 {
-	str2ba(device->address, bdaddr);
+	bacpy(bdaddr, &device->bdaddr);
 }
 
 const gchar *device_get_path(struct btd_device *device)
@@ -1155,6 +1248,46 @@ uint8_t device_get_auth(struct btd_device *device)
 	return device->auth;
 }
 
+static gboolean start_discovery(gpointer user_data)
+{
+	struct btd_device *device = user_data;
+
+	device_browse(device, NULL, NULL, NULL);
+
+	device->discov_timer = 0;
+
+	return FALSE;
+}
+
+gboolean device_set_paired(DBusConnection *conn, struct btd_device *device,
+			struct bonding_request_info *bonding)
+{
+	dbus_bool_t paired = TRUE;
+
+	device_set_temporary(device, FALSE);
+
+	dbus_connection_emit_property_changed(conn, device->path,
+						DEVICE_INTERFACE, "Paired",
+						DBUS_TYPE_BOOLEAN, &paired);
+
+	if (device->discov_timer)
+		return FALSE;
+
+	/* If we were initiators start service discovery immediately.
+	 * However if the other end was the initator wait a few seconds
+	 * before SDP. This is due to potential IOP issues if the other
+	 * end starts doing SDP at the same time as us */
+	if (bonding)
+		device_browse(device, bonding->conn,
+				bonding->msg, NULL);
+	else
+		device->discov_timer = g_timeout_add(DISCOVERY_TIMER,
+							start_discovery,
+							device);
+
+	return TRUE;
+}
+
 int btd_register_device_driver(struct btd_device_driver *driver)
 {
 	device_drivers = g_slist_append(device_drivers, driver);
